#include <windows.h>
#include <stdio.h>

#include "vmmc.h"
#define DBGPRINT(l, s) printf s

int iterations = 10000;
int spinCount;
int num_words = 4;
int num_pages = 1;
int nbytes;
int nwords, start_words;

int32 *dstBuf;
int32 *srcBuf;
int32 *srcbufend;
int32 *dstbufend;
int32 *proxy;
unsigned int rbufId = 2;
int isSlave=0;

void *
pageAlignedMalloc(int npages)
{
  void *block;
  void *ret;
  int i, PS = vmmc_WordSize() * vmmc_PageSize(), tmp;

  block = (void *)malloc((npages + 2) * PS);
  if (block == NULL) {
    fprintf(stderr, "page aligned malloc failed\n");
    exit(1);
  }
  /* touch each page to make sure they are allocated */
  for (i = 0; i < npages + 2; i++)
    tmp = ((long *)block)[i * vmmc_PageSize()];
  /* align to page beginning */
  ret = (void *)(PS * (1 + ((long)block / PS)));
  ((void **)ret)[-1] = block;
  return ret;
}

void
main(int argc, char *argv[])
{
  vmmc_pid_t		remotePid;
  vmmc_node_t		*hids, remoteHostId, myHostId;
  vmmc_exphandle_t	exphandle;
  vmmc_imphandle_t	imphandle;

  char *slargv[10];
  int errc, nhosts;
	LARGE_INTEGER	start, stop, freq;
  double t;
  int i;

//	SET_DEBUG_LEVEL(0);
  if (argc < 4) {
    printf("%s iterations start_words end_words\n", argv[0]);
    exit(0);
  }

	num_words  = atoi(argv[3]);
	iterations = atoi(argv[1]);
  start_words = atoi(argv[2]);
  
  num_pages = (num_words + vmmc_PageSize()) / vmmc_PageSize();
  nbytes = num_words * sizeof(int32);
  
	errc = vmmc_Parent(&remoteHostId, &remotePid);
	if (errc < 0) {
		vmmc_Error(errc, "getting parent id");
		exit(1);
	}

	if (remotePid == 0) { /* the master */
    isSlave=0;
	  /* print machine configuration */
	  errc = vmmc_SessionHosts(&nhosts, &hids);
	  if (errc < 0) {
		  vmmc_Error(errc, "getting hosts");
		  exit(1);
	  }
	  for (i = 0; i < nhosts; i++) {
		  DBGPRINT(0, ("\thost[%2d] : %s\n", i, vmmc_NodeToName(hids[i])));
	  }

	  /*
	   *	select a remote host
	   */
	  myHostId = vmmc_MyNode();
	  if (nhosts > 1) 
		  if (vmmc_EqualNodes(hids[0], myHostId)) 
		    remoteHostId = hids[1];
	    else 
			  remoteHostId = hids[0];
	  else 
		  remoteHostId = hids[0];
	  
    /* prepare slave args */
    slargv[0] = argv[1];
    slargv[1] = argv[2];
    slargv[2] = argv[3];
    slargv[3] = NULL;

	  DBGPRINT(0,
		  ("%s: spawning slave on node %s\n",
			argv[0], vmmc_NodeToName(remoteHostId)));

	  errc = vmmc_Spawn(argv[0], slargv, remoteHostId, &remotePid);
	  if (errc < 0) {
		  vmmc_Error(errc, "spawning slaves");
		  exit(1);
	  }

	  DBGPRINT(0, ("%s: slave spawned, slavePid= %u on %s\n",
	         argv[0], remotePid, vmmc_NodeToName(remoteHostId)));
  }
  else { /* slave */
    isSlave=1;
    printf("%s: parentPid= %u parentNode= %s\n",
	         argv[0], remotePid, vmmc_NodeToName(remoteHostId));
  }

  srcBuf = pageAlignedMalloc(num_pages);
  dstBuf = pageAlignedMalloc(num_pages);
  for(i=0; i<num_words; i++) {
    dstBuf[i] = 0;
    srcBuf[i] = 0;
  }

  errc = vmmc_ExportRecvBuf(dstBuf, num_words, rbufId, NULL, 0, 0, &exphandle);
  if (errc < 0) {
    vmmc_Error(errc, "exporting local recv buf");
    exit(1);
  }

  errc = vmmc_ImportRecvBuf(remoteHostId, remotePid, rbufId, &imphandle, &proxy, &nwords);
  if (errc < 0) {
    vmmc_Error(errc, "importing remote recv buf");
    exit(1);
  }

  printf("%s: successful import: nwords=%d, proxy=0x%x, dstbuf=0x%x\n", 
         argv[0], nwords, (int)proxy, (int)dstBuf);

  /* send once to establish translations */
  if ((errc = vmmc_SendData(srcBuf, proxy, num_words)) < 0) {
    vmmc_Error(errc, "send msg to remote");
    exit(1);
  }

	QueryPerformanceFrequency(&freq);

  for(; start_words<=num_words; start_words+=start_words) {
    srcbufend = &(srcBuf[start_words - 1]);
    dstbufend = &(dstBuf[start_words - 1]);
    nbytes = start_words * sizeof(int32);

		QueryPerformanceCounter(&start);

    if (isSlave) { /* slave: wait and send */
      for (i = 1; i <= iterations; i++) {
        while (*dstbufend != i)
          spinCount++;
        *srcbufend = i;
        if ((errc = vmmc_SendData(srcBuf, proxy, start_words)) < 0) {
          vmmc_Error(errc, "send msg to master");
          exit(1);
        }
      }
    }
    else { /* master: send and wait */
      for (i = 1; i <= iterations; i++) {
        *srcbufend = i;
        if ((errc = vmmc_SendData(srcBuf, proxy, start_words)) < 0) {
          vmmc_Error(errc, "send msg to slave");
          exit(1);
        }
        while (*dstbufend != i)
          spinCount++;
      }
    }

		QueryPerformanceCounter(&stop);

	  t = (double)(stop.QuadPart - start.QuadPart)/freq.QuadPart*1000000.0;

    printf("niter=%7d nbytes=%7d OneWayLat=%7.3f usec Band=%7.3f MB/s\n",
	         iterations,
	         nbytes,
	         t/2.0/iterations,
	         2.0*nbytes*iterations/t
	     );
    fflush(stdout);
  }

	if (! isSlave) {
		DBGPRINT(0, ("\n\n                       VMMC Rocks!!\n"));
		DBGPRINT(0, ("\n\n%s: please press the return key to exit ...\n", argv[0]));
		getchar();
	}
	DBGPRINT(0, ("%s: test finished!\n", argv[0]));
}
