#include <malloc.h>
#include "sbl_support.h"

#define START_SPAWN_NODE 1
#define SLAVE_FLAG "--slave"
#define SQUID_BUF_ID 10000
#define PROC_INDEX_BUF_ID (SQUID_BUF_ID+1)
#define SQUID_ACK_ID_BEGIN (SQUID_BUF_ID+2)

static BOOL is_slave=0; 
static vmmc_node_t *nodeid_info=NULL;
static volatile vmmc_pid_t *squid_info=NULL;
static unsigned total_procs=0;
static unsigned my_proc_index=0;

BOOL IsSlave()
{
  return is_slave;
}

BOOL IsMaster()
{
  return !is_slave;
}

vmmc_node_t GetNode(unsigned index)
{
  assert(index<total_procs);
  return nodeid_info[index];
}

vmmc_pid_t GetSquid(unsigned index)
{
  assert(index<total_procs);
  return squid_info[index];
}

unsigned GetMyProcIndex()
{
  return my_proc_index;
}

unsigned TotalProcs()
{
  return total_procs;
}

BOOL IsProc(unsigned index)
{
  return my_proc_index==index;
}

void CheckSlaveArg(unsigned *argc, char **argv)
{
  unsigned i;

  if (*argc>=2 && strcmp(argv[1], SLAVE_FLAG)==0) {
    /* this is a slave */
    is_slave=1;
    /* shift the arguments */
    (*argc)--;
    for(i=1; i<*argc; i++)
      argv[i]=argv[i+1];
  }
  else { /* print out the augument */
    fprintf(stderr, "\n start executing...\n");
    for (i=0; i<*argc; i++)
      fprintf(stderr, "%s ", argv[i]);
    fprintf(stderr, "\n");
    is_slave=0;
  }
}
#if 0
BOOL SpawnSlavesOnly(unsigned slaves, unsigned argc, char **argv)
{
  int i;
  int node;
  int skipped;

  /* allocate process info table */
  total_procs=slaves+1;
  nodeid_info=SafeMalloc(sizeof(vmmc_node_t)*total_procs,
			 "Cannot allocate process info table\n");
  squid_info=SafeMalloc(sizeof(vmmc_pid_t)*total_procs, 
			"Cannot allocate squid info table\n");

  /* get hosts info */
  errc=vmmc_SessionHosts(&nhosts, &hids);
  if (errc<0) {
    vmmc_Error(errc, "Geting hosts info");
    exit(1);
  }

  /* get master info */
  if (IsSlave()) {
    /* this is a slave */
    errc=vmmc_Parent(&nodeid_info[0], &squid_info[0]);
    if (errc<0) {
      vmmc_Error(errc, "Getting parent id");
      exit(1);
    }
  }
  else { /* this is a master */
    nodeid_info[0]=vmmc_MyNode();
    squid_info[0]=vmmc_MyPid();
  }

  /* select nodes to be involved */
  node=0;
  i=1;
  skipped=0;
#if 1
  while(node<nhosts && (hids[node].s_addr!=nodeid_info[0].s_addr))
    node++;
#endif
  while(i<=slaves && node<nhosts) {
    if (hids[node].s_addr==nodeid_info[0].s_addr && !skipped) {
      /* skip master node */
      node++;
      skipped=1;
      continue;
    }
    /* spawn a process */
    nodeid_info[i]=hids[node];
    i++; node++;
  }
  if (i<=slaves) {
    fprintf(stderr, "Only has %d nodes, cannot spawn %d slaves\n", node, slaves);
    exit(1);
  }
    
  /* master spawns slaves */
  if (IsMaster()) {
    char **sargv;
    char *exec_path;

    /* prepare slave argument */
    sargv=(char**)SafeMalloc(sizeof(char *)*(argc+1),
			 "Cannot allocate process info table\n");
    sargv[0]=SLAVE_FLAG;
    for(i=1; i<argc; i++)
      sargv[i]=argv[i];
    sargv[argc]=NULL;

    /* get the executable absolute path */
#ifdef WIN32
#define PATH_MAX MAXPATHLEN
    if (*argv[0]=='/')
      exec_path=argv[0];
    else {
      exec_path=_getcwd(NULL, PATH_MAX+1);
      if (strlen(exec_path)+strlen(argv[0])+1>PATH_MAX) {
	fprintf(stderr, "Absolute executable name is too long\n");
	exit(1);
      }
      strcat(exec_path, "/");
      strcat(exec_path, argv[0]);
    }
   
#else
    if (*argv[0]=='/')
      exec_path=argv[0];
    else {
      exec_path=getcwd(NULL, PATH_MAX+1);
      if (strlen(exec_path)+strlen(argv[0])+1>PATH_MAX) {
	fprintf(stderr, "Absolute executable name is too long\n");
	exit(1);
      }
      strcat(exec_path, "/");
      strcat(exec_path, argv[0]);
    }
#endif
    /* spawn processes */
    for(i=1; i<=slaves; i++) {
      errc=vmmc_Spawn(exec_path, sargv, nodeid_info[i], &(squid_info[i]));
      if (errc<0) {
	vmmc_Error(errc, "Spawning slaves");
	exit(1);
      }
      else {
	fprintf(stderr, "Slave %d spawn successfully on %s: squid=%u\n", 
		i, vmmc_NodeToName(nodeid_info[i]), squid_info[i]);
      }
    }

    /* clean up */
    if (exec_path!=argv[0])
      free(exec_path);
    free(sargv);
  }
  
  /* check vmmc_WordSize */
  if (vmmc_WordSize()%sizeof(int32)!=0) {
    fprintf(stderr, "vmmc_WordSize is not a mutiple of size of int32\n");
    exit(1);
  }
  if (sizeof(vmmc_pid_t)%vmmc_WordSize()!=0) {
    fprintf(stderr, "Size of vmmc_pid_t is not a mutiple of vmmc_WordSize\n");
    exit(1);
  }
}
#endif
BOOL SpawnSlaves(unsigned slaves, unsigned argc, char **argv)
{
  /* Spawn Slaves create slaves, init the Proc table for all 
   * slaves by exchanging messages, adjust argc and argv for slaves
   */ 

  int errc;
  unsigned nhosts;
  vmmc_node_t *hids;
  unsigned node;
  unsigned i;
  BOOL skipped=0;
  /* allocate process info table */
  total_procs=slaves+1;
  nodeid_info=SafeMalloc(sizeof(vmmc_node_t)*total_procs,
			 "Cannot allocate process info table\n");
  squid_info=SafeMalloc(sizeof(vmmc_pid_t)*total_procs, 
			"Cannot allocate squid info table\n");

  /* get hosts info */
  errc=vmmc_SessionHosts(&nhosts, &hids);
  if (errc<0) {
    vmmc_Error(errc, "Geting hosts info");
    exit(1);
  }

  /* get master info */
  if (IsSlave()) {
    /* this is a slave */
    errc=vmmc_Parent(&nodeid_info[0], &squid_info[0]);
    if (errc<0) {
      vmmc_Error(errc, "Getting parent id");
      exit(1);
    }
  }
  else { /* this is a master */
    nodeid_info[0]=vmmc_MyNode();
    squid_info[0]=vmmc_MyPid();
  }

  /* select nodes to be involved */
  node=0;
  i=1;
  skipped=0;
#if 1
  while(node<nhosts && !vmmc_EqualNodes(hids[node], nodeid_info[0]))
    node++;
#endif
  while(i<=slaves && node<nhosts) {
    if (vmmc_EqualNodes(hids[node], nodeid_info[0]) && !skipped) {
      /* skip master node */
      node++;
      skipped=1;
      continue;
    }
    /* spawn a process */
    nodeid_info[i]=hids[node];
    i++; node++;
  }
  if (i<=slaves) {
    fprintf(stderr, "Only has %d nodes, cannot spawn %d slaves\n", node, slaves);
    exit(1);
  }
    
  /* master spawns slaves */
  if (IsMaster()) {
    char **sargv;
    char *exec_path;

    /* prepare slave argument */
    sargv=(char**)SafeMalloc(sizeof(char *)*(argc+1),
			 "Cannot allocate process info table\n");
    sargv[0]=SLAVE_FLAG;
    for(i=1; i<argc; i++)
      sargv[i]=argv[i];
    sargv[argc]=NULL;

    /* get the executable absolute path */
#ifdef WIN32
#define PATH_MAX _MAX_PATH
    if (*argv[0]=='/')
      exec_path=argv[0];
    else {
      exec_path=_getcwd(NULL, PATH_MAX+1);
      if (strlen(exec_path)+strlen(argv[0])+1>PATH_MAX) {
	fprintf(stderr, "Absolute executable name is too long\n");
	exit(1);
      }
      strcat(exec_path, "/");
      strcat(exec_path, argv[0]);
    }
   
#else
    if (*argv[0]=='/')
      exec_path=argv[0];
    else {
      exec_path=getcwd(NULL, PATH_MAX+1);
      if (strlen(exec_path)+strlen(argv[0])+1>PATH_MAX) {
	fprintf(stderr, "Absolute executable name is too long\n");
	exit(1);
      }
      strcat(exec_path, "/");
      strcat(exec_path, argv[0]);
    }
#endif
    
    /* spawn processes */
    for(i=1; i<=slaves; i++) {
      errc=vmmc_Spawn(exec_path, sargv, nodeid_info[i], &(squid_info[i]));
      if (errc<0) {
	vmmc_Error(errc, "Spawning slaves");
	exit(1);
      }
      else {
	fprintf(stderr, "Slave %d spawn successfully on %s: squid=%u\n", 
		i, vmmc_NodeToName(nodeid_info[i]), squid_info[i]);
      }
    }

    /* clean up */
    if (exec_path!=argv[0])
      free(exec_path);
    free(sargv);
  }
  
  /* check vmmc_WordSize */
  if (vmmc_WordSize()%sizeof(int32)!=0) {
    fprintf(stderr, "vmmc_WordSize is not a mutiple of size of int32\n");
    exit(1);
  }
  if (sizeof(vmmc_pid_t)%vmmc_WordSize()!=0) {
    fprintf(stderr, "Size of vmmc_pid_t is not a mutiple of vmmc_WordSize\n");
    exit(1);
  }

  /* master and slaves exchange out all procs info to all procs */
  if (IsMaster()) { /* for master */
    volatile int32 *ack_array;
    vmmc_exphandle_t exphandle;
    vmmc_imphandle_t imphandle1, imphandle2;

    ack_array=(int32 *)SafeMalloc(sizeof(int32)*total_procs,
				  "Cannot allocate ack array\n");
    my_proc_index=0;
    
    fprintf(stderr, "Master: exchanging proc_info table: "); 
    /* master send squid_info to slave one by one */
    for(i=1; i<=slaves; i++) {
      int32 *proxy_squid_info, *proxy_proc_index;
      unsigned nwords, nwords_proc_index;

      /* Master export ack for slave */
      ack_array[i]=0;
      errc=vmmc_ExportRecvBuf(ack_array+i, sizeof(int32)/vmmc_WordSize(),
			     SQUID_ACK_ID_BEGIN+i, NULL, 0, 0, &exphandle);
      if (errc<0) {
	vmmc_Error(errc, "Master exporting ack");
	exit(1);
      }

      /* Master import squid_info recv buffer of slave */
      errc=vmmc_ImportRecvBuf(nodeid_info[i], squid_info[i], SQUID_BUF_ID,
			     &imphandle1, &proxy_squid_info, &nwords);
      if (errc<0) {
	vmmc_Error(errc, "Master importing squid_info");
	exit(1);
      }
      /* check if the size matches expected */ 
      if (sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs!=nwords) {
	/* this happened */
	fprintf(stderr, "Master: squid_info size imported is %i but should be %i\n", 
		nwords, sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs);
	exit(1);
      }
      assert(proxy_squid_info!=NULL);

      /* Master import proc_index of slave */
      errc=vmmc_ImportRecvBuf(nodeid_info[i], squid_info[i], PROC_INDEX_BUF_ID,
			     &imphandle1, &proxy_proc_index, &nwords_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master importing proc_index");
	exit(1);
      }
      /* check if the size matches expected */ 
      if (sizeof(int32)/vmmc_WordSize()!=nwords_proc_index) {
	fprintf(stderr, "Master: proc_index size imported is %i but should be %i\n", 
		nwords_proc_index, sizeof(int32)/vmmc_WordSize());
	exit(1);
      }
      assert(proxy_proc_index!=NULL);
      
      /* send master's squid_info */
      errc=vmmc_SendData(squid_info, proxy_squid_info, nwords);
      if (errc<0) {
	vmmc_Error(errc, "Master sending squid_info");
	exit(1);
      }

      /* send proc_index */
      vmmc_SendData(&i, proxy_proc_index, nwords_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master sending proc_index");
	exit(1);
      }
      
      /* wait for ack */
      while(ack_array[i]==0);
      if ((unsigned)(ack_array[i])!=i) {
	fprintf(stderr, "Slave index is not returned in ack\n");
	exit(1);
      }
      
      /* clean up */
#ifndef UNIMPORT_NOT_IMPLEMENTED
      /* unimport proxy_squid_info */
      errc=vmmc_UnimportRecvBuf(proxy_squid_info);
      if (errc<0) {
	vmmc_Error(errc, "Master unimporting squid_info");
	exit(1);
      }

      /* unimport proxy_proc_index */
      errc=vmmc_UnimportRecvBuf(proxy_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master unimporting proc_index");
	exit(1);
      }
#endif
#ifndef UNEXPORT_NOT_IMPLEMENTED
      /* unexport squid_ack */
      errc=vmmc_UnexportRecvBuf(SQUID_ACK_ID_BEGIN+i);
      if (errc<0) {
	vmmc_Error(errc, "Master unexporting ack");
	exit(1);
      }
#endif
      fprintf(stderr, ".");
    }
    free((void *)ack_array);
    fprintf(stderr, " done\n");
  }
  else { /* for slave */
    vmmc_exphandle_t exphandle1, exphandle2;
    vmmc_imphandle_t imphandle;
    volatile int32 recved=0;
    vmmc_pid_t parent_squid=squid_info[0];
    int32 *proxy_ack;
    unsigned nwords_ack;

    /* export squid_info buf for recv */
    errc=vmmc_ExportRecvBuf(squid_info, sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs,
			   SQUID_BUF_ID, NULL, 0, 0, &exphandle1);
    if (errc<0) {
      vmmc_Error(errc, "Slave exporting squid_info");
      exit(1);
    }

    /* export the recv for proc_index */
    errc=vmmc_ExportRecvBuf(&recved, sizeof(int32)/vmmc_WordSize(),
			   PROC_INDEX_BUF_ID, NULL, 0, 0, &exphandle2);
    if (errc<0) {
      vmmc_Error(errc, "Slave exporting recv for proc_index");
      exit(1);
    }

    /* wait for recv for my_proc_index and then clear */
    while(recved==0);
    my_proc_index=recved;
    recved=0;

    /* check if master's view is different from slave's view */
    if (squid_info[0]!=parent_squid || 
	my_proc_index==0 ||
	squid_info[my_proc_index]!=vmmc_MyPid()) {
      fprintf(stderr, "Master's squid is inconsistent in slave's view\n");
      exit(1);
    }

    /* import master's ack flag */
    errc=vmmc_ImportRecvBuf(nodeid_info[0], squid_info[0], SQUID_ACK_ID_BEGIN+my_proc_index,
			   &imphandle, &proxy_ack, &nwords_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave importing master's ack");
      exit(1);
    }
    /* check if the size matches expected */ 
    if (sizeof(int32)/vmmc_WordSize()!=nwords_ack) {
      fprintf(stderr, "Slave: ack size imported is %i but should be %i\n", 
	      nwords_ack, sizeof(int32)/vmmc_WordSize());
      exit(1);
    }
    assert(proxy_ack!=NULL);

    /* send back ack */
    vmmc_SendData(&my_proc_index, proxy_ack, nwords_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave sending back ack");
      exit(1);
    }

    /* clean up */
#ifndef UNIMPORT_NOT_IMPLEMENTED
    /* unimport ack */
    errc=vmmc_UnimportRecvBuf(proxy_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave unimporting proxy_ack");
      exit(1);
    }
#endif
#ifndef UNEXPORT_NOT_IMPLEMENTED
    /* unexport squid_info */
    errc=vmmc_UnexportRecvBuf(SQUID_BUF_ID);
    if (errc<0) {
      vmmc_Error(errc, "Slave unexporting  squid_info");
      exit(1);
    }
    
    /* unexport squid_info recv flag */
    errc=vmmc_UnexportRecvBuf(PROC_INDEX_BUF_ID);
    if (errc<0) {
      vmmc_Error(errc, "Slave unexporting recv for proc_index");
      exit(1);
    }
#endif    
  }
  return 1;
}




void *PageAlignedMalloc(unsigned pages)
{
  unsigned page_size=vmmc_WordSize()*vmmc_PageSize();
  void *block, *ret;
  volatile int tmp=0;
  unsigned i;

  block=malloc((pages+2)*page_size);
  if (block==NULL) {
    fprintf(stderr, "PageAlignedMalloc: Fail to allocate %d pages\n", pages);
    exit(1);
  }
  ret=(void *)(((unsigned)block/page_size+2)*page_size);
  ((void **)ret)[-1]=block;

  /* touch each page */
  for(i=0; i<pages; i++)
    ((int32 *)ret)[i*vmmc_PageSize()]=i;
  return ret;
}

BOOL PageAlignedFree(void *buf)
{
  void *block=((void**)buf)[-1];
  free(block);
  return 1;
}

void *NonOverlapMalloc(unsigned size)
{
  unsigned page_size=vmmc_WordSize()*vmmc_PageSize();
  void *block, *ret;
  volatile int tmp=0;
  unsigned i;

  assert(page_size==4096);

  block=malloc(size+2*page_size);
  if (block==NULL) {
    fprintf(stderr, "NonOverlapMalloc: Fail to allocate %d bytes\n", size);
    exit(1);
  }
  ret=(void *)((char *)block+page_size);
  ((void **)ret)[-1]=block;

  /* touch each page */
  for(i=0; i<size; i+=page_size)
    ((char*)ret)[i]=i;
  ((char*)ret)[size-1]=i;

  return ret;
}

BOOL NonOverlapFree(void *buf)
{
  void *block=((void**)buf)[-1];
  free(block);
  return 1;
}

void *SafeMalloc(unsigned size, const char *err_format, ...)
{
  va_list ap;
  unsigned page_size;
  unsigned i;
  volatile char c;

  void *p=malloc(size);
  if (p==NULL) {
    va_start(ap, err_format);
    vfprintf(stderr, err_format, ap);
    va_end(ap);
    exit(1);
  }
    
  /* touch one for each page */
  page_size=vmmc_PageSize()*vmmc_WordSize();
  c=0;
  for(i=0; i<size; i+=page_size) {
#if 1 /* touch for write */
    ((char*)p)[i]=0;
#else /* only read */
    c+=((char*)p)[i];
#endif
  }

  /* touch the last one */
#if 1 /* touch for write */
  ((char*)p)[size-1]=0;
#else /* only read */
  c+=((char*)p)[size-1];
#endif
    
  return p;
}

#if 0
BOOL ModifiedSpawnSlaves(unsigned slaves, unsigned argc, char **argv)
{
  /* Spawn Slaves create slaves, init the Proc table for all 
   * slaves by exchanging messages, adjust argc and argv for slaves
   */ 

  int errc;
  unsigned nhosts;
  vmmc_node_t *hids;
  unsigned node;
  unsigned i;
  BOOL skipped=0;

  /* allocate process info table */
  total_procs=slaves+1;
  nodeid_info=SafeMalloc(sizeof(vmmc_node_t)*total_procs,
			 "Cannot allocate process info table\n");
  squid_info=SafeMalloc(sizeof(vmmc_pid_t)*total_procs, 
			"Cannot allocate squid info table\n");

  /* get hosts info */
  errc=vmmc_SessionHosts(&nhosts, &hids);
  if (errc<0) {
    vmmc_Error(errc, "Geting hosts info");
    exit(1);
  }

  /* get master info */
  if (IsSlave()) {
    /* this is a slave */
    errc=vmmc_Parent(&nodeid_info[0], &squid_info[0]);
    if (errc<0) {
      vmmc_Error(errc, "Getting parent id");
      exit(1);
    }
  }
  else { /* this is a master */
    nodeid_info[0]=vmmc_MyNode();
    squid_info[0]=vmmc_MyPid();
  }

  /* select nodes to be involved */
  node=0;
  i=1;
  skipped=0;
#if 1
  while(node<nhosts && (hids[node].s_addr!=nodeid_info[0].s_addr))
    node++;
#endif
  while(i<=slaves && node<nhosts) {
    if (hids[node].s_addr==nodeid_info[0].s_addr && !skipped) {
      /* skip master node */
      node++;
      skipped=1;
      continue;
    }
    /* spawn a process */
    nodeid_info[i]=hids[node];
    i++; node++;
  }
  if (i<=slaves) {
    fprintf(stderr, "Only has %d nodes, cannot spawn %d slaves\n", node, slaves);
    exit(1);
  }
    
  /* master spawns slaves */
  if (IsMaster()) {
    char **sargv;
    char *exec_path;

    /* prepare slave argument */
    sargv=(char**)SafeMalloc(sizeof(char *)*(argc+1),
			 "Cannot allocate process info table\n");
    sargv[0]=SLAVE_FLAG;
    for(i=1; i<argc; i++)
      sargv[i]=argv[i];
    sargv[argc]=NULL;

    /* get the executable absolute path */
    if (*argv[0]=='/')
      exec_path=argv[0];
    else {
      exec_path=getcwd(NULL, PATH_MAX+1);
      if (strlen(exec_path)+strlen(argv[0])+1>PATH_MAX) {
	fprintf(stderr, "Absolute executable name is too long\n");
	exit(1);
      }
      strcat(exec_path, "/");
      strcat(exec_path, argv[0]);
    }
    
    /* spawn processes */
    for(i=1; i<=slaves; i++) {
      errc=vmmc_Spawn(exec_path, sargv, nodeid_info[i], &(squid_info[i]));
      if (errc<0) {
	vmmc_Error(errc, "Spawning slaves");
	exit(1);
      }
      else {
	fprintf(stderr, "Slave %d spawn successfully on %s: squid=%u\n", 
		i, vmmc_NodeToName(nodeid_info[i]), squid_info[i]);
      }
    }

    /* clean up */
    if (exec_path!=argv[0])
      free(exec_path);
    free(sargv);
  }
  
  /* check vmmc_WordSize */
  if (vmmc_WordSize()%sizeof(int32)!=0) {
    fprintf(stderr, "vmmc_WordSize is not a mutiple of size of int32\n");
    exit(1);
  }
  if (sizeof(vmmc_pid_t)%vmmc_WordSize()!=0) {
    fprintf(stderr, "Size of vmmc_pid_t is not a mutiple of vmmc_WordSize\n");
    exit(1);
  }

  /* master and slaves exchange out all procs info to all procs */
  if (IsMaster()) { /* for master */
    volatile int32 *ack_array;
    ack_array=(int32 *)SafeMalloc(sizeof(int32)*total_procs,
				  "Cannot allocate ack array\n");
    my_proc_index=0;
    
    fprintf(stderr, "Master: exchanging proc_info table: "); 
    /* master send squid_info to slave one by one */
    for(i=1; i<=slaves; i++) {
      int32 *proxy_squid_info, *proxy_proc_index;
      unsigned nwords, nwords_proc_index;

      /* Master export ack for slave */
#if 0
      ack_array[i]=0;
      errc=vmmc_ExportRecvBuf(ack_array+i, sizeof(int32)/vmmc_WordSize(),
			     SQUID_ACK_ID_BEGIN+i, NULL, 0);
      if (errc<0) {
	vmmc_Error(errc, "Master exporting ack");
	exit(1);
      }
#endif
      /* Master import squid_info recv buffer of slave */
      errc=vmmc_ImportRecvBuf(nodeid_info[i], squid_info[i], SQUID_BUF_ID,
			     &proxy_squid_info, &nwords);
      if (errc<0) {
	vmmc_Error(errc, "Master importing squid_info");
	exit(1);
      }
      /* check if the size matches expected */ 
      if (sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs!=nwords) {
	/* this happened */
	fprintf(stderr, "Master: squid_info size imported is %i but should be %i\n", 
		nwords, sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs);
	exit(1);
      }
      assert(proxy_squid_info!=NULL);

      /* send master's squid_info */
      errc=vmmc_SendData(squid_info+1, proxy_squid_info+1, nwords-1);
      if (errc<0) {
	vmmc_Error(errc, "Master sending squid_info");
	exit(1);
      }

      errc=vmmc_SendData(squid_info, proxy_squid_info, 1);
      if (errc<0) {
	vmmc_Error(errc, "Master sending squid_info[0]");
	exit(1);
      }

      /* Master import proc_index of slave */
      errc=vmmc_ImportRecvBuf(nodeid_info[i], squid_info[i], PROC_INDEX_BUF_ID,
			     &proxy_proc_index, &nwords_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master importing proc_index");
	exit(1);
      }
      /* check if the size matches expected */ 
      if (sizeof(int32)/vmmc_WordSize()!=nwords_proc_index) {
	fprintf(stderr, "Master: proc_index size imported is %i but should be %i\n", 
		nwords_proc_index, sizeof(int32)/vmmc_WordSize());
	exit(1);
      }
      assert(proxy_proc_index!=NULL);
      
      /* send proc_index */
      vmmc_SendData(&i, proxy_proc_index, nwords_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master sending proc_index");
	exit(1);
      }
#if 0      
      /* wait for ack */
      while(ack_array[i]==0);
      if (ack_array[i]!=i) {
	fprintf(stderr, "Slave index is not returned in ack\n");
	exit(1);
      }
#endif
      
      /* clean up */
#ifndef UNIMPORT_NOT_IMPLEMENTED
      /* unimport proxy_squid_info */
      errc=vmmc_UnimportRecvBuf(proxy_squid_info);
      if (errc<0) {
	vmmc_Error(errc, "Master unimporting squid_info");
	exit(1);
      }

      /* unimport proxy_proc_index */
      errc=vmmc_UnimportRecvBuf(proxy_proc_index);
      if (errc<0) {
	vmmc_Error(errc, "Master unimporting proc_index");
	exit(1);
      }
#endif
#ifndef UNEXPORT_NOT_IMPLEMENTED
      /* unexport squid_ack */
      errc=vmmc_UnexportRecvBuf(SQUID_ACK_ID_BEGIN+i);
      if (errc<0) {
	vmmc_Error(errc, "Master unexporting ack");
	exit(1);
      }
#endif
      fprintf(stderr, ".");
    }
    free((void*)ack_array);
    fprintf(stderr, " done\n");
  }
  else { /* for slave */
    volatile int32 recved=0;
    vmmc_pid_t parent_squid=squid_info[0];
#if 0
    int32 *proxy_ack;
    unsigned nwords_ack;
#endif

    squid_info[0]=0;
    /* export squid_info buf for recv */
    errc=vmmc_ExportRecvBuf(squid_info, sizeof(vmmc_pid_t)/vmmc_WordSize()*total_procs,
			   SQUID_BUF_ID, NULL, 0);
    if (errc<0) {
      vmmc_Error(errc, "Slave exporting squid_info");
      exit(1);
    }

    /* wait for squid_info[0] */
    while(squid_info[0]==0);    

    /* export the recv for proc_index */
    errc=vmmc_ExportRecvBuf(&recved, sizeof(int32)/vmmc_WordSize(),
			   PROC_INDEX_BUF_ID, NULL, 0);
    if (errc<0) {
      vmmc_Error(errc, "Slave exporting recv for proc_index");
      exit(1);
    }

    /* wait for recv for my_proc_index and then clear */
    while(recved==0);
    my_proc_index=recved;
    recved=0;

    /* check if master's view is different from slave's view */
    if (squid_info[0]!=parent_squid || 
	my_proc_index==0 ||
	squid_info[my_proc_index]!=vmmc_MyPid()) {
      fprintf(stderr, "Master's squid is inconsistent in slave's view\n");
      exit(1);
    }

#if 0
    /* import master's ack flag */
    errc=vmmc_ImportRecvBuf(nodeid_info[0], squid_info[0], SQUID_ACK_ID_BEGIN+my_proc_index,
			   &proxy_ack, &nwords_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave importing master's ack");
      exit(1);
    }
    /* check if the size matches expected */ 
    if (sizeof(int32)/vmmc_WordSize()!=nwords_ack) {
      fprintf(stderr, "Slave: ack size imported is %i but should be %i\n", 
	      nwords_ack, sizeof(int32)/vmmc_WordSize());
      exit(1);
    }
    assert(proxy_ack!=NULL);

    /* send back ack */
    vmmc_SendData(&my_proc_index, proxy_ack, nwords_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave sending back ack");
      exit(1);
    }
#endif
    /* clean up */
#ifndef UNIMPORT_NOT_IMPLEMENTED
    /* unimport ack */
    errc=vmmc_UnimportRecvBuf(proxy_ack);
    if (errc<0) {
      vmmc_Error(errc, "Slave unimporting proxy_ack");
      exit(1);
    }
#endif
#ifndef UNEXPORT_NOT_IMPLEMENTED
    /* unexport squid_info */
    errc=vmmc_UnexportRecvBuf(SQUID_BUF_ID);
    if (errc<0) {
      vmmc_Error(errc, "Slave unexporting  squid_info");
      exit(1);
    }
    
    /* unexport squid_info recv flag */
    errc=vmmc_UnexportRecvBuf(PROC_INDEX_BUF_ID);
    if (errc<0) {
      vmmc_Error(errc, "Slave unexporting recv for proc_index");
      exit(1);
    }
#endif    
  }
  return 1;
}
#endif