/* cycle.c
 *         by Xiang Yu
 *
 * General timing facilities for Pentium, with cycle precision.
 */

#include "timing.h"

static int initialized=0;
static unsigned cpu_clock;
static float us_per_cycle, ms_per_cycle, s_per_cycle;

unsigned GetCPUClock()
{
  if (!initialized)
    TimingInit();
  return cpu_clock;
}

#ifdef WIN32
Cycle GetCycle()
{
  volatile ULARGE_INTEGER t;
  __asm {
	  __asm _emit 0x0f;
	  __asm _emit 0x31;
	  __asm mov t.LowPart, eax;
	  __asm mov t.HighPart, edx;
  }
  return t.QuadPart;
}

#else
Cycle GetCycle()
{
  Cycle t;
  asm volatile (".byte 0x0f, 0x31" : "=A" (t));
  return t;
}
#endif

float Interval(Cycle start, Cycle end)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)(end-start))*s_per_cycle;
}

float mInterval(Cycle start, Cycle end)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)(end-start))*ms_per_cycle;
}

float uInterval(Cycle start, Cycle end)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)(end-start))*us_per_cycle;
}

float Period(Cycle c)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)c)*s_per_cycle;
}

float mPeriod(Cycle c)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)c)*ms_per_cycle;
}

float uPeriod(Cycle c)
{
  if (!initialized)
    TimingInit();
  return ((float)(SCycle)c)*us_per_cycle;
}

void TimingInit()
{
  Cycle s,e;
  
  if (initialized) 
    return;

  /* mesure the system clock frequency in MHz */
  GETCYCLE(s);
#ifdef WIN32
  Sleep(2000); /* in ms */
#else
  sleep(2);
#endif
  GETCYCLE(e);  

  cpu_clock=(unsigned)((e-s+1000000)/2000000);
  fprintf(stderr,"CPU CLOCK is %uMHz\n",cpu_clock);
  us_per_cycle=1.0f/cpu_clock;
  ms_per_cycle=1.0e-3f/cpu_clock;
  s_per_cycle=1.0e-6f/cpu_clock;
  initialized=1;
}
