/*****************************************************************************
 *
 *							mbulk.c
 *
 *		This program demonstrates bulk message pingpong and checks the
 *	validaty of received data.
 *
 *****************************************************************************/
#include <stdio.h>
#include <windows.h>
#include "vmmc.h"


#define	MBULK_TEST_BUFFER_ID		2

unsigned long	TotalErrorCount;



void prepare_data(uint32* buf, unsigned int iter, unsigned int nwords)
{
  unsigned int i, j;

  for (i = 0, j = 2 * iter; i < nwords; i++) {
    buf[i] = j + i;
  }
}

void check_data(uint32* buf, unsigned int iter, unsigned int nwords)
{
  unsigned int i, j;

  while (buf[nwords - 1] != ((2 * iter) + nwords - 1))
	  Sleep(0);
  for (i = 0, j = 2 * iter; i < nwords; i++) {
    if (buf[i] != j + i) {
      printf("DATA ERROR ***** bad value %d should be %d\n",
			buf[i], j + i);
      ++TotalErrorCount;
    }
  }
}

void test_mbulk(vmmc_node_t remoteNode, vmmc_pid_t remotePid,
				unsigned int nwords, unsigned int iterations,
				int IamMaster)
{
  vmmc_result_t		errc;			/* error code */
  unsigned int		i = 0;
  vmmc_exphandle_t	exphandle;
  vmmc_imphandle_t	imphandle;
  uint32*			dstBuf;
  uint32*			srcBuf;
  int32 *			proxy;

  printf("mbulk test started ... ...\n");

  dstBuf = (int32*) VirtualAlloc(NULL, nwords*sizeof(int32),
								MEM_COMMIT, PAGE_READWRITE);
  srcBuf = (int32*) VirtualAlloc(NULL, nwords*sizeof(int32),
								MEM_COMMIT, PAGE_READWRITE);
  errc = vmmc_ExportRecvBuf(
					dstBuf,
					nwords,
					MBULK_TEST_BUFFER_ID,
					NULL,
					0,
					0,
					&exphandle);
  if (errc != vmmc_Success) {
    vmmc_Error(errc, "exporting slave buf");
    exit(1);
  }
  errc = vmmc_ImportRecvBuf(
						remoteNode,
						remotePid,
						MBULK_TEST_BUFFER_ID,
						&imphandle, &proxy, &nwords);
  if (errc != vmmc_Success) {
    vmmc_Error(errc, "importing master buf");
    exit(1);
  }
  printf("mbulk: successful import: nwords=%d,proxy=0x%x, dstBuf=0x%x\n",
		nwords, (int)proxy, (int)dstBuf);

  for (i = 1; i <= iterations;) {
	  if (IamMaster) {
		  check_data(dstBuf, i, nwords);
	  }
	  
	  prepare_data(srcBuf, i, nwords);
	  errc = vmmc_SendData(srcBuf, proxy, nwords);
	  if (errc != vmmc_Success ) {
			vmmc_Error(errc, "send msg to master");
			exit(1);
	  }
	  
	  srcBuf[nwords - 1] = 0xFFFFFFFF;
	  if ( ! IamMaster) {
		  check_data(dstBuf, i, nwords);
	  }
	  printf("mbulk: msg number %d received\n", i);
	  i++;
  }
  
  printf("mbulk: Exiting...DATA ERRORS = %d\n", TotalErrorCount);
}





void
main(int argc, char *argv[])
{
  vmmc_result_t		errc;
  unsigned int		i;
  char				*slargv[3];
  uint32			nhosts;
  vmmc_node_t		*nodelist, remoteNode;
  vmmc_pid_t		remotePid;
  unsigned int		nwords, iterations;

  if ( argc != 3 ) {
	  fprintf(stderr, "usage: %s nwords iterations\n", argv[0]);
	  exit(1);
  }

  nwords     = atoi(argv[1]);
  iterations = atoi(argv[2]);

  errc = vmmc_Parent(&remoteNode, &remotePid);
  if ( errc != vmmcErr_NoParent) {
    test_mbulk(remoteNode, remotePid, nwords, iterations, 0);
    exit(0);
  }

  /* print machine configuration */
  errc = vmmc_SessionHosts(&nhosts, &nodelist);
  if (errc != vmmc_Success) {
    vmmc_Error(errc, "getting hosts");
    exit(1);
  }
  printf("mbulk: number of hosts = %d\n", nhosts);
  for (i = 0; i < nhosts; i++) {
    printf("mbulk: hostname[%2d] = %s\n", vmmc_NodeToName(nodelist[i]));
  }

  /* prepare slave args */
  slargv[0] = argv[1];
  slargv[1] = argv[2];
  slargv[2] = NULL;

  if (nhosts > 1) {
    if ( vmmc_EqualNodes(nodelist[0], vmmc_MyNode()) ) {
      remoteNode = nodelist[1];
    } else {
      remoteNode = nodelist[0];
    }
  } else {
    remoteNode = nodelist[0];
  }
  printf("mbulk: spawning slave on %s\n", vmmc_NodeToName(remoteNode));

  errc = vmmc_Spawn(argv[0], slargv, remoteNode, &remotePid);
  if (errc != vmmc_Success) {
    vmmc_Error(errc, "spawning slaves");
    exit(1);
  }
  printf("mbulk: slave spawned, pid=%u on %s\n",
		remotePid, vmmc_NodeToName(remoteNode));

  test_mbulk(remoteNode, remotePid, nwords, iterations, 1);
}
