/*****************************************************************************
 *
 *							notif.c
 *
 *		This program demonstrates the notification facility in VMMC interface.
 *	It uses the old SBL.H. Users of VMMC are encouraged to use the new VMMC
 *	interface, specified in vmmc.h
 *
 *****************************************************************************/
#include <windows.h>
#include <stdio.h>
#include "sbl.h"


#define	ITERATIONS	10000

unsigned int rbufId = 2;
Int32 rbuf, i=0;
Int32 *rbufProxy;
int nwords;
unsigned long t1, t2;
volatile int flag=-1;

char *SlaveName  = "Slave(" __FILE__ ")";
char *MasterName = "Master(" __FILE__ ")";



void
MsgHandler(Int32 * last_word, Int32 word)
{
  flag = *last_word;
  // printf("got a notification %d at 0x%x\n", (Int32)last_word, word);
}

void
Slave(void)
{
  int errc;			/* error code */
  SBL_Squid parentSquid;
  NodeId hid;			/* host id */

  errc = SBL_Parent(&hid, &parentSquid);
  if (errc < 0) {
    SBL_Error(errc, "getting parent id");
    exit(1);
  }

  rbuf = 0;
  errc = SBL_ExportRecvBuf(&rbuf, 1, rbufId, MsgHandler, 0);
  if (errc < 0) {
    SBL_Error(errc, "exporting recv buf");
    exit(1);
  }

  errc = SBL_ImportRecvBuf(hid, parentSquid, rbufId, &rbufProxy, &nwords);
  if (errc < 0) {
    SBL_Error(errc, "importing master buf");
    exit(1);
  }
  fprintf(stderr, "%s Successful import, nwords = %d\n", "notif:", nwords);

  i=0;
  errc = SBL_SendMsgWithNotify(&i, rbufProxy, 1);
  if (errc < 0) {
    	SBL_Error(errc, "send msg to master");
    	exit(1);
  }
  for (i=1; i<=ITERATIONS; i++) {
  	while (((volatile)flag) != i) ;
    errc = SBL_SendMsgWithNotify(&i, rbufProxy, 1);
    if (errc < 0) {
    	  SBL_Error(errc, "send msg to master");
    	  exit(1);
    }
  }
  fprintf(stderr, "%s Exiting...\n", "notif:");
}

void
main(int argc, char *argv[])
{
  SBL_Squid slaveSquid;
  int errc;
  int i;
  char *slargv[2];
  int nhosts;
  NodeId *hids, hid;

  if ((argc == 2) && (strcmp(argv[1], "slave") == 0)) {
    Slave();
    exit(0);
  }
  fprintf(stderr, "MsgHandler = 0x%x\n", (int)&MsgHandler);

  /* master */
  if (argc > 1) {		/* master fromat */
    fprintf(stderr, "bad format: %s takes no arguments\n", argv[0]);
    exit(1);
  }
  /* print machine configuration */
  errc = SBL_Hosts(&nhosts, &hids);
  if (errc < 0) {
    SBL_Error(errc, "getting hosts");
    exit(1);
  }
  fprintf(stderr, "%s Num hosts = %d\n", "notif:", nhosts);
  for (i = 0; i < nhosts; i++) {
    fprintf(stderr, "%s Host number %d, host address %s\n",
	    "notif:", i, SBL_GetHostName(*(hids + i)));
  }

  /* prepare slave args */
  slargv[0] = "slave";
  slargv[1] = NULL;

  fprintf(stderr, "%s spawning slave\n", "notif:");
  hid = SBL_GetMyNodeId();
  if (nhosts > 1) {
    if (SBL_EqualNodeId(&hid, &hids[0]) )
      hid = hids[1];
    else
      hid = hids[0];
  }


  errc = SBL_Spawn(argv[0], slargv, hid, &slaveSquid);
  if (errc < 0) {
    SBL_Error(errc, "spawning slaves");
    exit(1);
  }
  fprintf(stderr,
	  "%s Slave spawned, slaveSquid = %u, slave node = %s\n",
	  "notif:", slaveSquid, SBL_GetHostName(hid));

  rbuf = 0;
  errc = SBL_ExportRecvBuf(&rbuf, 1, rbufId, MsgHandler, 0);
  if (errc < 0) {
    SBL_Error(errc, "exporting recv buf");
    exit(1);
  }
  fprintf(stderr, "%s Successful export\n", "notif:");

  errc = SBL_ImportRecvBuf(hid, slaveSquid, rbufId, &rbufProxy, &nwords);
  if (errc < 0) {
    SBL_Error(errc, "importing master buf");
    exit(1);
  }
  fprintf(stderr, "%s Successful import, nwords = %d\n", "notif:", nwords);

 	while (((volatile)flag) != 0) ;

  t1 = GetTickCount();
  for (i=1; i<=ITERATIONS; i++) {
  	errc = SBL_SendMsgWithNotify(&i, rbufProxy, 1);
  	if (errc < 0) {
    		SBL_Error(errc, "send msg to master");
    		exit(1);
  	}
  	while (((volatile)flag) != i) ;
  }
  t2 = GetTickCount();

  printf("Elapsed time for %d pingpongs is %ld ms\n", ITERATIONS, (t2-t1));
  printf("Round-trip notification latency is %d us\n",
	(t2-t1)*1000/ITERATIONS);
}
