/*****************************************************************************
 *
 *							spawn.c
 *
 *		This program demonstrates the vmmc_Spawn interface.
 *
 *****************************************************************************/
#include <stdio.h>
#include <windows.h>
#include "vmmc.h"
#include "utils.h"

#define	MAX_LEVEL	2

char *SlaveName  = "Slave(" __FILE__ ")";
char *MasterName = "Master(" __FILE__ ")";
char MyName[256];

void main(int argc, char *argv[])
{
  int				level = 1;
  int				isMaster=0;
  vmmc_result_t		errc;
  uint32			i;
  char			 *	slargv[2];
  uint32			nhosts;
  vmmc_pid_t		parentPid;
  vmmc_pid_t		remotePid;
  vmmc_node_t		parentNode;			/* host id */
  vmmc_node_t	*	completeNodeList;
  vmmc_node_t	*	availableNodeList;
  char				numberBuffer[32];

  errc = vmmc_Parent(&parentNode, &parentPid);

  if ( errc == vmmcErr_NoParent) {
	isMaster = 1;
	sprintf(MyName, "Master-%s ", MasterName);
  } else {
	sprintf(MyName, "Slave-%s ", SlaveName);
  }

  if ( isMaster) {
    if (argc>1) {
      level = atoi(argv[1]);
    }
  } else {
    if ( argc != 2) {
      fprintf(stderr, "%s: expect a single level argument\n", MyName);
    }
    level = atoi(argv[1]);
  }

  if (level > MAX_LEVEL) {
     fprintf(stderr, "%s: level (%d) too huge, exiting\n", MyName, level);
     exit(1);
  }

  printf("%s: master started: spawning all children for %d levels\n",
	MyName, level);

  errc = vmmc_AllHosts(&nhosts, &completeNodeList);
  if (errc< 0) {
    fprintf(stderr, "%s: vmmc_AllHosts returned %d\n", MyName, errc);
    exit(errc);
  }

  printf("%s: listing all nodes in the system ...\n", MyName);
  for (i=0; i<nhosts; i++) {
    printf("%s: system_host[%2d] = %s  nodeid=%d\n",
		MyName, i,
		vmmc_NodeToName(completeNodeList[i]),
		vmmc_NameToNode(vmmc_NodeToName(completeNodeList[i])).addr);
  }

  errc = vmmc_SessionHosts(&nhosts, &availableNodeList);
  if (errc< 0) {
    fprintf(stderr, "%s: vmmc_Hosts returned %d\n", MyName, errc);
    exit(errc);
  }

  if (nhosts == 0) {
    printf("%s: listing all nodes in my session ...NONE\n", MyName);
    exit(0);
  } else {
    printf("%s: listing all nodes in my session ...\n", MyName);
  }
  for (i=0; i<nhosts; i++) {
    printf("%s: my host[%2d] = %s\n",
	MyName, i, vmmc_NodeToName(availableNodeList[i]));
  }


  if (level == 0) {
    printf("%s: exiting at the bottom level\n", MyName);
    exit(0);
  }

  printf("%s: spawning %d processes on remote nodes ...\n", MyName, nhosts);
  sprintf(numberBuffer, "%d", level-1);
  slargv[0] = numberBuffer;
  slargv[1] = NULL;
  for (i=0; i<nhosts; i++) {
     printf("%s: spawning a process on node %s\n",
	   MyName, vmmc_NodeToName(availableNodeList[i]));
   errc = vmmc_Spawn(argv[0], slargv, availableNodeList[i], &remotePid);
    if (errc < 0) {
		fprintf(stderr, "spawn: vmmc_Spawn failed: %s\n",
				vmmc_ErrorStr(errc));
      exit(1);
    }
    printf("%s: spawned a process (squid=%d) on node %s\n",
	   MyName, remotePid, vmmc_NodeToName(availableNodeList[i]));
  }
  printf("%s: spawned %d processes on remote nodes.\n", MyName, nhosts);
}
