/*****************************************************************************
 *
 *	sbl.h
 *
 *  by Xiang Yu, Princeton University, Feb, 1999
 *
 *****************************************************************************
 *	wrapper header file for old SBL programs to run on VMMC release version
 *  WITHOUT source modification
 *****************************************************************************/

#ifndef __SBL_H__
#define __SBL_H__

#include "vmmc.h"

#define  SBL_Ok                       vmmc_Success
#define  SBL_ENoSuchError             vmmcErr_NoSuchError
#define  SBL_ENoMem                   vmmcErr_NoMem
#define  SBL_ENoSuchNode	          vmmcErr_NoSuchNode
#define  SBL_ENoSuchProcess           vmmcErr_NoSuchProcess
#define  SBL_ENoPhysDestSpace         vmmcErr_NoPhysDestSpace
#define  SBL_ENoVirtDestSpace         vmmcErr_NoVirtDestSpace
#define  SBL_EBadProxyExtension       vmmcErr_BadProxyExtension
#define  SBL_ESending                 vmmcErr_InProgress
#define  SBL_EOverlappedBufs          vmmcErr_OverlappedBufs
#define  SBL_EBadRbufState            vmmcErr_BadRbufState
#define  SBL_EBadIbufState            vmmcErr_BadIbufState
#define  SBL_EBadAlignment            vmmcErr_BadAlignment
#define  SBL_EBadSize                 vmmcErr_BadSize
#define  SBL_EBadProxyAddr            vmmcErr_BadProxyAddr
#define  SBL_EBlockedNotifications    vmmcErr_BlockedNotifications
#define  SBL_EUnblockedNotifications  vmmcErr_UnblockedNotifications
#define  SBL_ENotInHandler            vmmcErr_NotInHandler
#define  SBL_EBadArg                  vmmcErr_BadArg
#define  SBL_ENotImplemented          vmmcErr_NotImplemented
#define  SBL_EReceiving               vmmcErr_InProgress

/* SBL_Export flags */

#define EXP_BUF_REDIRECTABLE BUF_REDIRECTABLE
#define EXP_BUF_GLOBAL_SPACE BUF_GLOBAL_SPACE
#define EXP_BUF_FETCH_BUF_FLAG BUF_READWRITE

typedef int32	 	Int32;
typedef uint32	UInt32;
typedef vmmc_callback_t     SBL_MsgHandler;
typedef vmmc_exphandle_t    SBL_BufHandle_t;
typedef vmmc_result_t	      SBL_Status_t;
typedef vmmc_perm_t         SBL_Permissions_t;
typedef vmmc_node_t         NodeId;
typedef vmmc_pid_t          SBL_Squid;

#define SBL_InvalidSquid    vmmc_InvalidPid

/*** Functions **********************************************/

#define CALLTYPE CALLSTYLE
#ifdef __cplusplus
extern "C" {
#endif

/* mapping setup functions */
SBL_Status_t CALLTYPE
SBL_ExportRecvBuf(
			Int32 * 			buf,		/* word-aligned local buffer*/
			unsigned int		nwords,		/* size of the buffer in words*/
			unsigned int		bufid,		/* assigned by application*/
			SBL_MsgHandler		handler,	/* NULL if no notification */
			int					eserved
	);


SBL_Status_t CALLTYPE 
SBL_Export(
			Int32 *				buf,
			unsigned int		nwords,
			unsigned int		bufid,
			SBL_MsgHandler		handler,
			int					flags,	/* -redirectableBuf, contiguous bits */
			SBL_Permissions_t	perm,	/* used to match export-import reqs */
			SBL_BufHandle_t *	bhandle	/* OUT */
	);

SBL_Status_t CALLTYPE
SBL_ImportRecvBuf(
	                NodeId			node,		/* remote node's id */
	                SBL_Squid		squid,		/* remote process' id */
	                unsigned int	bufid,		/* remote buffer's id */
	                Int32 **		proxyAddr,	/* filled by this call*/
	                int *			nwords		/* filled by this call */
                  );

/* returns impId or Error  */
int CALLTYPE 
SBL_ImportRecvBufAsync(
	                     NodeId node,
       	               SBL_Squid squid,
       	               unsigned int bufid
                      );

SBL_Status_t CALLTYPE 
SBL_ImportRecvBufStatus(
	                      int	impId,
	                      Int32 **proxyAddr,
	                      int *nwords
                       );

SBL_Status_t CALLTYPE
SBL_UnimportRecvBuf(Int32 * 		proxyAddr);

SBL_Status_t CALLTYPE
SBL_UnexportRecvBuf(unsigned int	bufid);

int CALLTYPE SBL_Importers(
			                            unsigned int	bufid,
		                              int 			*numImporters,
		                              SBL_Squid **	impSquids,
		                              NodeId **		impHosts
                                 );

/******************* data transfer functions *******************/
#define SBL_SendMsg vmmc_SendData
#define SBL_SendMsgWithNotify vmmc_SendDataNotify
#define SBL_GetMsg vmmc_GetData

int CALLTYPE
SBL_SendMsgAsync(
		             Int32 *localSrcAddr,
			           Int32 *remoteDestProxyAddr,
			           int	 nwords
                );

int CALLTYPE
SBL_SendMsgAsyncWithNotify(
		             Int32 *localSrcAddr,
			           Int32 *remoteDestProxyAddr,
			           int	 nwords
                );

SBL_Status_t CALLTYPE
SBL_AsyncSendStatus(int msgid);

int CALLTYPE
SBL_GetMsgAsync(
			          Int32 *localDstBuffer, 
	              Int32 *remoteSrcProxyAddr, 
			          unsigned int	nwords
               );

SBL_Status_t CALLTYPE
SBL_AsyncGetStatus(int msgid);

/******************* notification functions *******************/
#define SBL_BlockNotifications vmmc_BlockNotifications
#define SBL_UnblockNotifications vmmc_UnblockNotifications

/******************* cluster remote functions *******************/
#define SBL_Spawn vmmc_Spawn

SBL_Status_t CALLTYPE 
SBL_Parent(NodeId * parentNode, SBL_Squid * parentSquid);

#define SBL_Hosts vmmc_SessionHosts
#define SBL_AllHosts vmmc_AllHosts
#define SBL_GetSquid vmmc_MyPid
#define SBL_GetMyNodeId vmmc_MyNode
#define SBL_GetNodeId vmmc_NameToNode
#define SBL_GetHostName vmmc_NodeToName


int CALLTYPE
SBL_EqualNodeId(NodeId *node1, NodeId *node2);

int CALLTYPE
SBL_GetLogicalNodeId(NodeId nodeid);

/******************* redirection functions *******************/
#define SBL_MsgEnd vmmc_DataEnd
#define SBL_ClearMsgEnd vmmc_ClearDataEnd
#define SBL_PostRedir vmmc_PostRedir
#define SBL_CancelRedir vmmc_EndRedir


/******************* miscellaneous functions *******************/
#define SBL_Error vmmc_Error
#define SBL_Version vmmc_Version
#define SBL_PageSize vmmc_PageSize
#define SBL_WordSize vmmc_WordSize

/*	Flexible Debugging */
#define SBL_SetDebugLevel vmmc_SetDebugLevel

#ifdef NI_LOCKS
/* NI lock functions */
void CALLTYPE vmmc_InitLock(int num_nodes);
int CALLTYPE vmmc_LockIsLocal(int lock_num, int lhome, int *time);
int CALLTYPE vmmc_AcquireLock(int lock_num, int lhome, int *time);
int CALLTYPE vmmc_ReleaseLock(int lock_num, int lhome, int *time);

#define SBL_InitLock vmmc_InitLock
#define SBL_LockIsLocal vmmc_LockIsLocal
#define SBL_AcquireLock vmmc_AcquireLock
#define SBL_ReleaseLock vmmc_ReleaseLock
#endif /* NI_LOCKS */

#ifdef	__cplusplus
}
#endif	__cplusplus


#endif	__SBL_H__